# Copyright 2020-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MONGODB-AWS authentication support for PyMongo."""

import os
from functools import wraps

from base64 import standard_b64encode
from datetime import tzinfo, timedelta, datetime


import boto3
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials
from botocore.utils import parse_to_aware_datetime

from pymongo_auth_aws.errors import PyMongoAuthAwsError


"""MONGODB-AWS credentials."""
class AwsCredential:
    def __init__(self, username, password, token, refresh_needed=None):
        self.username = username
        self.password = password
        self.token = token
        self.refresh_needed = refresh_needed


_credential_buffer_seconds = 60
__cached_credentials = None
__use_cached_credentials = None


def get_use_cached_credentials():
    """Determine whether we are using cached credentials."""
    return __use_cached_credentials


def set_use_cached_credentials(value):
    """Set whether we are using cached credentials."""
    global __use_cached_credentials
    __use_cached_credentials = value


def get_cached_credentials():
    """Central point for accessing cached credentials."""
    global __cached_credentials
    creds = __cached_credentials
    if creds and creds.refresh_needed is not None:
        if creds.refresh_needed(_credential_buffer_seconds):
            creds = __cached_credentials = None
    return creds


def set_cached_credentials(credentials):
    """Central point for setting cached credentials."""
    global __cached_credentials
    if __use_cached_credentials:
        __cached_credentials = credentials


ZERO = timedelta(0)

# A Python 2.7-compliant UTC class (from stdlib docs).
# When we drop Python 2.7 support we can use `timezone.utc` instead.

class _UTC(tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO

utc = _UTC()


def aws_temp_credentials():
    """Construct temporary MONGODB-AWS credentials."""
    # Store the variable locally for safe threaded access.
    creds = get_cached_credentials()
    if creds:
        return creds

    try:
        session = boto3.Session()
        creds = session.get_credentials()
        # Use frozen credentials to prevent a race condition if there
        # is a refresh between property accesses.
        frozen = creds.get_frozen_credentials()
    except Exception:
        # If temporary credentials cannot be obtained then drivers MUST
        # fail authentication and raise an error.
        set_cached_credentials(None)
        raise PyMongoAuthAwsError(
            'temporary MONGODB-AWS credentials could not be obtained')

    # The botocore Credentials object does not expose the expiration
    # directly, instead we use the refresh_needed method to determine
    # whether the credentials are expired.
    refresh_needed = getattr(creds, 'refresh_needed', None)
    creds = AwsCredential(
        frozen.access_key, frozen.secret_key, frozen.token, refresh_needed
    )
    # Only cache credentials that need to be refreshed from
    # an external source.
    if refresh_needed is not None:
        set_cached_credentials(creds)
    return creds


_AWS4_HMAC_SHA256 = 'AWS4-HMAC-SHA256'
_AWS_SERVICE = 'sts'


def _get_region(sts_host):
    """Return the AWS region to use for the given host."""
    # Drivers must also validate that the host is greater than 0 and
    # less than or equal to 255 bytes per RFC 1035.
    if not sts_host or len(sts_host) > 255:
        raise PyMongoAuthAwsError(
            "Server returned an invalid sts host: %s" % (sts_host,))

    parts = sts_host.split('.')
    if len(parts) == 1 or sts_host == 'sts.amazonaws.com':
        return 'us-east-1'  # Default

    # Check for empty labels (eg "invalid..host" or ".invalid.host").
    if not all(parts):
        raise PyMongoAuthAwsError(
            "Server returned an invalid sts host: %s" % (sts_host,))

    return parts[1]


def _aws_auth_header(credentials, server_nonce, sts_host):
    """Signature Version 4 Signing Process to construct the authorization header
    """
    region = _get_region(sts_host)

    request_parameters = 'Action=GetCallerIdentity&Version=2011-06-15'
    encoded_nonce = standard_b64encode(server_nonce).decode('utf8')
    request_headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Content-Length': str(len(request_parameters)),
        'Host': sts_host,
        'X-MongoDB-Server-Nonce': encoded_nonce,
        'X-MongoDB-GS2-CB-Flag': 'n',
    }
    request = AWSRequest(method="POST", url="/", data=request_parameters,
                         headers=request_headers)
    boto_creds = Credentials(credentials.username, credentials.password,
                             token=credentials.token)
    auth = SigV4Auth(boto_creds, "sts", region)
    auth.add_auth(request)
    final = {
        'a': request.headers['Authorization'],
        'd': request.headers['X-Amz-Date']
    }
    if credentials.token:
        final['t'] = credentials.token
    return final


def _handle_credentials(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception:
            set_cached_credentials(None)
            raise
    return inner


class AwsSaslContext(object):
    """MONGODB-AWS authentication support.

    :Parameters:
      - `credentials`: The :class:`AwsCredential` to use for authentication.
    """
    def __init__(self, credentials):
        self._credentials = credentials
        self._step = 0
        self._client_nonce = None

    def step(self, server_payload):
        """Step through the SASL conversation.

        :Parameters:
          - `server_payload`: The server payload (SASL challenge). Must be a
            bytes-like object.

        :Returns:
          The response payload for the next SASL step.

        :Raises:
          :class:`~pymongo_auth_aws.PyMongoAuthAwsError` on error.
        """
        self._step += 1
        if self._step == 1:
            return self._first_payload()
        elif self._step == 2:
            return self._second_payload(server_payload)
        else:
            raise PyMongoAuthAwsError('MONGODB-AWS failed: too many steps')
        pass

    @_handle_credentials
    def _first_payload(self):
        """Return the first SASL payload."""
        # If a username and password are not provided, drivers MUST query
        # a link-local AWS address for temporary credentials.
        if self._credentials.username is None:
            self._credentials = aws_temp_credentials()

        # Client first.
        client_nonce = os.urandom(32)
        self._client_nonce = client_nonce
        payload = {'r': self.binary_type()(client_nonce), 'p': 110}
        return self.binary_type()(self.bson_encode(payload))

    @_handle_credentials
    def _second_payload(self, server_payload):
        """Return the second and final SASL payload."""
        if not server_payload:
            raise PyMongoAuthAwsError(
                'MONGODB-AWS failed: server payload empty')

        server_payload = self.bson_decode(server_payload)
        server_nonce = server_payload['s']
        if len(server_nonce) != 64 or not server_nonce.startswith(
                self._client_nonce):
            raise PyMongoAuthAwsError("Server returned an invalid nonce.")

        sts_host = server_payload['h']
        payload = _aws_auth_header(self._credentials, server_nonce, sts_host)
        return self.binary_type()(self.bson_encode(payload))

    # Dependency injection:
    def binary_type(self):
        """Return the bson.binary.Binary type."""
        raise NotImplementedError

    def bson_encode(self, doc):
        """Encode a dictionary to BSON."""
        raise NotImplementedError

    def bson_decode(self, data):
        """Decode BSON to a dictionary."""
        raise NotImplementedError



# Copyright 2020-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the pymongo-auth-aws module."""

from datetime import datetime, timedelta
import threading
import json
import os
import sys
import tempfile

try:
    from http.server import  SimpleHTTPRequestHandler
    import socketserver
except ImportError:  # python 2
    from SimpleHTTPServer  import SimpleHTTPRequestHandler
    import SocketServer as socketserver

sys.path[0:0] = [""]

import botocore.session
from botocore.stub import Stubber

import bson
from bson.binary import Binary
import pymongo_auth_aws

from pymongo_auth_aws import auth
from pymongo_auth_aws.auth import _get_region, aws_temp_credentials, AwsSaslContext, AwsCredential
from pymongo_auth_aws.errors import PyMongoAuthAwsError

from test import unittest

# Ensure we are not using a local credentials file.
os.environ['AWS_SHARED_CREDENTIALS_FILE'] = '/tmp'
AWS_DATE_FORMAT = r"%Y-%m-%dT%H:%M:%SZ"
PORT = 8000
URI = 'http://localhost:%s' % PORT
RESPONSE = None


class MockHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        global RESPONSE
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        output = json.dumps(RESPONSE)
        self.wfile.write(output.encode('utf8'))


class TestAuthAws(unittest.TestCase):

    def setUp(self):
        auth.set_cached_credentials(None)
        os.environ.pop('AWS_CONTAINER_CREDENTIALS_FULL_URI', None)
        os.environ.pop('AWS_ACCESS_KEY_ID', None)
        os.environ.pop('AWS_SECRET_ACCESS_KEY', None)
        return unittest.TestCase.setUp(self)

    @classmethod
    def setUpClass(cls):
        cls.httpd = socketserver.TCPServer(("", PORT), MockHandler)
        cls.thread = threading.Thread(target=cls.httpd.serve_forever)
        cls.thread.setDaemon(False)
        cls.thread.start()

    @classmethod
    def tearDownClass(cls):
        cls.httpd.shutdown()
        cls.thread.join()

    def assertVersionLike(self, version):
        self.assertTrue(isinstance(version, str), msg=version)
        # There should be at least one dot: "1.0" or "1.0.0" not "1".
        self.assertGreaterEqual(len(version.split('.')), 2, msg=version)

    def test_version(self):
        self.assertVersionLike(pymongo_auth_aws.__version__)

    def test_region(self):
        # Default region is us-east-1.
        self.assertEqual('us-east-1', _get_region('sts.amazonaws.com'))
        self.assertEqual('us-east-1', _get_region('first'))
        self.assertEqual('us-east-1', _get_region('f'))
        # Otherwise, the region is the second label.
        self.assertEqual('second', _get_region('first.second'))
        self.assertEqual('second', _get_region('first.second.third'))
        self.assertEqual('second', _get_region('sts.second.amazonaws.com'))
        # Assert invalid hosts cause an error.
        self.assertRaises(PyMongoAuthAwsError, _get_region, '')
        self.assertRaises(PyMongoAuthAwsError, _get_region, 'i'*256)
        self.assertRaises(PyMongoAuthAwsError, _get_region, 'first..second')
        self.assertRaises(PyMongoAuthAwsError, _get_region, '.first.second')
        self.assertRaises(PyMongoAuthAwsError, _get_region, 'first.second.')

    def ensure_equal(self, creds, expected):
        self.assertEqual(creds.username, expected['AccessKeyId'])
        self.assertEqual(creds.password, expected['SecretAccessKey'])
        if 'SessionToken' in expected:
            self.assertEqual(creds.token, expected['SessionToken'])
        else:
            self.assertEqual(creds.token, expected['Token'])

    def test_aws_temp_credentials_env_variables(self):
        os.environ['AWS_ACCESS_KEY_ID'] = 'foo'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'bar'
        creds = aws_temp_credentials()
        del os.environ['AWS_ACCESS_KEY_ID']
        del os.environ['AWS_SECRET_ACCESS_KEY']
        self.assertEqual(creds.username, 'foo')
        self.assertEqual(creds.password, 'bar')
        self.assertEqual(creds.token, None)

    def test_aws_temp_credentials(self):
        global RESPONSE
        os.environ['AWS_CONTAINER_CREDENTIALS_FULL_URI'] = URI
        expected = dict(AccessKeyId='foo', SecretAccessKey='bar', Token='fizz', Expiration='2050-03-15T00:05:07Z')
        RESPONSE = expected
        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)

    def test_cache_credentials(self):
        global RESPONSE
        auth.set_use_cached_credentials(True)
        os.environ['AWS_CONTAINER_CREDENTIALS_FULL_URI'] = URI
        tomorrow = datetime.now(auth.utc) + timedelta(days=1)
        expected = dict(AccessKeyId='foo', SecretAccessKey='bar', Token='fizz', Expiration=tomorrow.strftime(AWS_DATE_FORMAT))
        RESPONSE = expected
        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)

        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)

    def test_local_creds_not_cached(self):
        os.environ['AWS_ACCESS_KEY_ID'] = 'foo'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'bar'
        creds = aws_temp_credentials()
        self.assertEqual(creds.username, 'foo')
        self.assertEqual(creds.password, 'bar')
        self.assertEqual(creds.token, None)

        os.environ['AWS_ACCESS_KEY_ID'] = 'fizz'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'buzz'
        creds = aws_temp_credentials()
        del os.environ['AWS_ACCESS_KEY_ID']
        del os.environ['AWS_SECRET_ACCESS_KEY']
        self.assertEqual(creds.username, 'fizz')
        self.assertEqual(creds.password, 'buzz')
        self.assertEqual(creds.token, None)

    def test_caching_disabled(self):
        global RESPONSE
        auth.set_use_cached_credentials(False)
        os.environ['AWS_CONTAINER_CREDENTIALS_FULL_URI'] = URI
        soon = datetime.now(auth.utc) + timedelta(minutes=10)
        expected = dict(AccessKeyId='foo', SecretAccessKey='bar', Token='fizz', Expiration=soon.strftime(AWS_DATE_FORMAT))
        RESPONSE = expected
        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)

        tomorrow = datetime.now(auth.utc) + timedelta(days=1)
        expected['Expiration'] = tomorrow.strftime(AWS_DATE_FORMAT)
        RESPONSE = expected
        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)

    def test_cache_expired(self):
        global RESPONSE
        auth.set_use_cached_credentials(True)
        os.environ['AWS_CONTAINER_CREDENTIALS_FULL_URI'] = URI
        expired = datetime.now(auth.utc) - timedelta(hours=1)
        expected = dict(AccessKeyId='foo', SecretAccessKey='bar', Token='fizz', Expiration=expired.strftime(AWS_DATE_FORMAT))
        RESPONSE = expected

        self.assertRaises(PyMongoAuthAwsError, aws_temp_credentials)

        expected['AccessKeyId'] = 'fizz'
        tomorrow = datetime.now(auth.utc) + timedelta(days=1)
        expected['Expiration'] = tomorrow.strftime(AWS_DATE_FORMAT)
        RESPONSE = expected
        creds = aws_temp_credentials()

        self.ensure_equal(creds, expected)

    def test_cache_expires_soon(self):
        global RESPONSE
        auth.set_use_cached_credentials(True)
        os.environ['AWS_CONTAINER_CREDENTIALS_FULL_URI'] = URI
        soon = datetime.now(auth.utc) + timedelta(seconds=30)
        expected = dict(AccessKeyId='foo', SecretAccessKey='bar', Token='fizz', Expiration=soon.strftime(AWS_DATE_FORMAT))
        RESPONSE = expected
        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)

        expected['AccessKeyId'] = 'fizz'
        tomorrow = datetime.now(auth.utc) + timedelta(days=1)
        expected['Expiration'] = tomorrow.strftime(AWS_DATE_FORMAT)
        RESPONSE = expected
        creds = aws_temp_credentials()
        self.ensure_equal(creds, expected)


class _AwsSaslContext(AwsSaslContext):

    def binary_type(self):
        """Return the bson.binary.Binary type."""
        return Binary

    def bson_encode(self, doc):
        """Encode a dictionary to BSON."""
        return bson.encode(doc)

    def bson_decode(self, data):
        """Decode BSON to a dictionary."""
        return bson.decode(data)



class TestAwsSaslContext(unittest.TestCase):

    def test_step(self):
        creds = AwsCredential('foo', 'bar', 'baz')
        test = _AwsSaslContext(creds)
        response = bson.decode(test.step(None))
        nonce = response['r'] + os.urandom(32)
        # Python 2.7 support.
        if sys.version_info[0] == 2:
            nonce = Binary(nonce)
        payload = bson.encode(dict(s=nonce, h='foo.com'))
        response = test.step(payload)
        self.assertIsInstance(response, Binary)


if __name__ == "__main__":
    unittest.main()



import os

from setuptools import setup, find_packages

with open('README.rst', 'rb') as f:
    LONG_DESCRIPTION = f.read().decode('utf8')

# Single source the version.
version_file = os.path.realpath(os.path.join(
    os.path.dirname(__file__), 'pymongo_auth_aws', 'version.py'))
version = {}
with open(version_file) as fp:
    exec(fp.read(), version)

setup(
    name="pymongo-auth-aws",
    version=version['__version__'],
    description="MONGODB-AWS authentication support for PyMongo",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=['test']),
    install_requires=['boto3', 'botocore'],
    extras_require={
        "test": ["pymongo"]
    },
    author="Shane Harvey",
    author_email="drivers-python-noreply@mongodb.com",
    url="https://github.com/mongodb/pymongo-auth-aws",
    keywords=["mongo", "mongodb", "pymongo-auth-aws", "pymongo", "MONGODB-AWS"],
    test_suite="test",
    license="Apache License, Version 2.0",
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database"]
)
